"""
stride.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

import numpy as np
import pandas as pd
import os
import re
import shutil
import subprocess


class Stride:
    """
    Store the results from stride.
    """

    # Initialize class instance
    def __init__(self, raw):
        """
        Initialize Stride instance.

        Parameters
        ----------
        raw : str
            Raw stride output.
        """

        # Initialize hidden variables
        self._raw = None
        self._data = None

        # Sets raw and creates processed data
        self.raw = raw

    # Slice data
    def __getitem__(self, item):
        return self._data[item]

    # Override __repr__
    def __repr__(self):
        return self._raw

    # Returns processed pandas DataFrame
    @property
    def data(self):
        """
        Raw data processed into pandas.DataFrame

        Returns
        -------
        pandas.DataFrame
            Processed data.
        """

        return self._data

    @data.setter
    def data(self, value):
        raise NotImplemented

    @property
    def raw(self):
        return self._raw

    @raw.setter
    def raw(self, raw):
        self._raw = raw

        # Filter for pertinent records in output
        records = re.sub(r'^(?!ASG).*$', '', raw, flags=re.MULTILINE).replace('\n\n', '\n').lstrip()

        # Sections of output
        sections = np.array([
            (3, 'record', '<U3'),
            (5, 'residue', '<U5'),
            (2, 'chain', '<U2'),
            (5, 'residue_id', 'int'),
            (5, 'residue_id2', 'int'),
            (5, 'secondary_structure', '<U4'),
            (14, 'secondary_structure_text', '<U14'),
            (10, 'phi', 'float'),
            (10, 'psi', 'float'),
            (10, 'area', 'float'),
            (10, 'extra', '<U10')
        ], dtype=[('width', 'i1'), ('column', '<U24'), ('type', '<U10')])

        # Parse records and return as DataFrame
        data = np.genfromtxt(records.split('\n'), delimiter=sections['width'], dtype=sections['type'], autostrip=True)
        data = pd.DataFrame(data.tolist(), columns=sections['column'])

        # Drop extraneous columns
        data = data.drop(columns=['record', 'residue_id2', 'extra'])

        # Return data
        self._data = data


# TODO it would be great to eventually bind stride to Python directly (as opposed to going through subprocess)
# Compute secondary structure with stride
def stride(fname, executable='stride'):
    """
    Compute the secondary structure using STRIDE

    In addition to the secondary structure, STRIDE outputs phi and psi dihedral angles as well as the solvent-accessible
    surface area.

    Parameters
    ----------
    fname : str
        Path to PDB file.
    executable : str
        Location of stride executable (Default: stride).

    Returns
    -------
    pandas.DataFrame
        DataFrame containing the results of STRIDE. Columns include: residue_id, chain, secondary_structure,
        secondary_structure_text, phi, psi, and area.

    Examples
    --------
    Computing the secondary structure as you would on the command line with STRIDE.
    >>> from molecular.external import stride
    >>> ss = stride('my.pdb')['secondary_structure'].values

    Computing the secondary structure on a Structure (or Trajectory)
    >>> from molecular.analysis import secondary_structure
    >>> from molecular.io import read_pdb
    >>> ss = secondary_structure(read_pdb('my.pdb'))
    """

    # Make sure executable exists
    if shutil.which('stride') is None and not os.path.exists(executable):
        raise AttributeError('executable %s not found. Download at http://webclu.bio.wzw.tum.de/stride/' % executable)
    
    # Run STRIDE and capture output
    process = subprocess.Popen([executable, fname], stdout=subprocess.PIPE)
    output, error = process.communicate()
    
    # Error check; make sure STRIDE finishes successfully
    if process.wait() != 0:
        raise SystemError('stride failed')

    # Return raw output
    return Stride(output.decode('ASCII'))



"""
Boltzmann
"""

import numpy as np

from scipy.constants import Boltzmann, Avogadro, atmosphere
import scipy.constants

AVOGADRO = scipy.constants.Avogadro
BOLTZMANN = scipy.constants.Boltzmann


# TODO make this easily usable.
class GasConstant:
    def __init__(self):
        pass

    @property
    def kcal_per_K_mol(self):
        return per_mole(joule_to_kcal(BOLTZMANN))


# Joule to kcal conversion
def joule_to_kcal(x):
    """
    Convert joule to kcal. 1 J = 1 / 4184. kcal

    Parameters
    ----------
    x : float
        Joule.

    Returns
    -------
    float
        kcal.
    """

    return x / 4184.


# kcal to Joule conversion
def kcal_to_joule(x):
    """
    Convert kcal to joule. 1 kcal = 4184 J.

    Parameters
    ----------
    x : float
        kcal.

    Returns
    -------
    float
        Joule.
    """

    return 4184. * x


def per_mole(x):
    """
    Convert quantity to per mole. Divide by NA.

    Parameters
    ----------
    x : float
        Quantity.

    Returns
    -------
    float
        Quantity per mole.
    """

    return x * AVOGADRO


# Boltzmann constant in kcal/K/mol
BOLTZMANN_KCAL_K_MOL = per_mole(joule_to_kcal(BOLTZMANN))
BOLTZMANN_KCAL_MOL_K = BOLTZMANN_KCAL_K_MOL
np.testing.assert_almost_equal(BOLTZMANN * AVOGADRO / 4184., BOLTZMANN_KCAL_K_MOL)

# Pressure
pressure = Avogadro * atmosphere / 1e30 / 4184.

if __name__ == '__main__':
    print(f'kb={BOLTZMANN_KCAL_MOL_K} kcal/mol')



"""
transform.py

language: Python
version: 3.7
author: C. Lockhart <chris@lockhartlab.org>
"""

from numba import njit
import numpy as np


def fit(a, b, backend='python'):
    """
    `a` and `b` must be paired.

    Parameters
    ----------
    a : molecular.Trajectory
    b : molecular.Trajectory
    backend

    Returns
    -------

    """

    # Number of structures
    n_structures = a.n_structures
    if n_structures != b.n_structures:
        raise AttributeError('a and b must have the same number of structures')

    # Number of atoms
    n_atoms = a.n_atoms
    if n_atoms != b.n_atoms:
        raise AttributeError('a and b must have the same number of atoms')

    # Number of dimensions
    n_dim = 3

    # Get coordinates
    a_xyz = a.xyz
    b_xyz = b.xyz

    # Centers
    a_xyz_center = np.tile(a_xyz.mean(axis=1), n_atoms).reshape(n_structures, n_atoms, n_dim)
    b_xyz_center = np.tile(b_xyz.mean(axis=1), n_atoms).reshape(n_structures, n_atoms, n_dim)

    # Move structures to center
    # a_xyz = to_origin(a.xyz)
    # b_xyz = to_origin(b.xyz)
    a_xyz = a_xyz - a_xyz_center
    b_xyz = b_xyz - b_xyz_center

    # Transpose b
    b_xyz_transpose = np.transpose(b_xyz, axes=[0, 2, 1])

    # Compute covariance matrix
    covariance_matrix = np.matmul(b_xyz_transpose, a_xyz)

    # Get rotation matrix
    rotation_matrix = _get_rotation_matrix(covariance_matrix)

    # Perform optimal rotation
    b_xyz = np.matmul(b_xyz, rotation_matrix)

    # Move b to a center
    b_xyz = b_xyz + a_xyz_center

    # Return
    return b_xyz


def to_origin(a):
    """

    Parameters
    ----------
    a

    Returns
    -------

    """

    # Compute the center
    # TODO put this in geometry.center ?
    center = a.xyz.mean(axis=1)

    # Transform
    a.xyz = a.xyz - center

    # Return
    return a


def _fit(a_xyz, b_xyz, backend='python'):
    """
    `a_xyz` and `b_xyz` must be paired.

    Parameters
    ----------
    a_xyz : numpy.ndarray
    b_xyz : numpy.ndarray
    backend : str

    Returns
    -------

    """

    # Extract dimensions
    n_structures, n_atoms, n_dim = a_xyz.shape

    # Sanity check
    if n_structures != b_xyz.shape[0]:
        raise AttributeError('a_xyz and b_xyz must have the same number of structures')
    if n_atoms != b_xyz.shape[1]:
        raise AttributeError('a_xyz and b_xyz must have the same number of atoms')

    # Centers
    a_xyz_center = np.tile(a_xyz.mean(axis=1), n_atoms).reshape(n_structures, n_atoms, n_dim)
    b_xyz_center = np.tile(b_xyz.mean(axis=1), n_atoms).reshape(n_structures, n_atoms, n_dim)

    # Move structures to center
    a_xyz = a_xyz - a_xyz_center
    b_xyz = b_xyz - b_xyz_center

    # Transpose b
    b_xyz_transpose = np.transpose(b_xyz, axes=[0, 2, 1])

    # Compute covariance matrix
    covariance_matrix = np.matmul(b_xyz_transpose, a_xyz)

    # Get rotation matrix
    rotation_matrix = _get_rotation_matrix(covariance_matrix)

    # Perform optimal rotation
    b_xyz = np.matmul(b_xyz, rotation_matrix)

    # Move b to a center
    b_xyz = b_xyz + a_xyz_center

    # Return
    return b_xyz



@njit
def _get_rotation_matrix(covariance_matrix):
    """
    Compute the rotation matrix from the method from Kabsch (1976) Acta Cryst. A32.



    Parameters
    ----------
    covariance_matrix

    Returns
    -------

    """
    n_structures = covariance_matrix.shape[0]
    rotation_matrix = []
    for i in range(n_structures):
        u, s, v = np.linalg.svd(covariance_matrix[i, :, :], full_matrices=True)
        s = np.diag(np.array([1., 1., np.sign(np.linalg.det(np.dot(u, v)))]))
        r = np.dot(np.dot(u, s), v)
        rotation_matrix.append(r)
    return rotation_matrix



from molecular.geometry import dihedral
import numpy as np


# TODO make this work with multiple chains
def phi(a):
    residue_ids = np.unique(a.topology['residue_id'])
    if not np.all(np.diff(residue_ids) == 1):
        raise AttributeError('residue_id must be sequential: %s' % residue_ids)

    n_structures = a.n_structures
    n_dim = a.n_dim

    # Skip over the first residue_id
    for residue_id in residue_ids[1:]:
        residue0 = a.select(residue_id=residue_id - 1)
        residue1 = a.select(residue_id=residue_id)

        atom0 = residue0.select(atom='C').xyz.reshape(n_structures, n_dim)
        atom1 = residue1.select(atom='N').xyz.reshape(n_structures, n_dim)
        atom2 = residue1.select(atom='CA').xyz.reshape(n_structures, n_dim)
        atom3 = residue1.select(atom='C').xyz.reshape(n_structures, n_dim)

        return dihedral(atom0, atom1, atom2, atom3)



"""
STRIDE----

ReadPDBFile
CheckChain

BackboneAngles
DiscrPhiPsi

PlaceHydrogens
FindHydrogenBonds
NoDoubleHBond

Helix
Sheet
BetaTurn
GammaTurn
Report"""

from . import backbone_dihedrals


def stride(trajectory):
    """

    Parameters
    ----------
    trajectory : Trajectory

    Returns
    -------

    """

    # Select protein
    protein = trajectory.query('protein')
    _check_protein_topology(protein.topology)

    # Compute backbone dihedral angles
    phi = backbone_dihedrals.phi(trajectory)
    psi = backbone_dihedrals.psi(trajectory)

    # Place hydrogen if necessary; let's skip with explicit solvent CHARMM structures

    # Compute hydrogen bonds


def _check_protein_topology(topology):
    """
    check chain step

    Parameters
    ----------
    a : Topology

    Returns
    -------

    """

    # Less than 5 residues?
    if topology.n_residues < 5:
        raise AttributeError('protein contains < 5 residues')


def _hydrogen_bonds(trajectory, distance_cutoff=6.):
    topology = trajectory.topology

    donors, hydrogens = _find_hydrogen_donors(topology)
    acceptors = _find_hydrogen_acceptors(topology)

    n_structures = trajectory.n_structures
    #TODO jit or cython or fortran
    distance = np.array([cdist(donors[i, :, :], acceptors[i, :, :]) for i in range(n_structures)])
    contact = distance <= distance_cutoff

    # energetic hydrogen bond


    # rose hydrogen bond

    # baker hydrogen bond

    #
    # if energetic bond and energy < 0 or rose or baker:
    #     flag as hydrogen bond


def _grid_energy(ca2, c, o, h, n):
    pass

def _identify_helix(trajectory):
    pass

def _find_hydrogen_donors(topology):
    """

    Parameters
    ----------
    trajectory

    Returns
    -------

    """

    donors = []
    hydrogens = []

    residue_ids = np.unique(topology['residue_id'])
    for i, residue_id in enumerate(residue_ids):
        residue = a.select(residue_id=residue_id)
        if i > 0:
            last_residue = a.select(residue_id=residue_id - 1)
            donors.append([
                _Atom(residue.select(atom='N')),
                _Atom(last_residue.select(atom='C')),
                _Atom(residue.select(atom='CA'))
            ])
        hydrogens.append(_Atom(residue.select(atom='H')))

        # there are specific residue hydrogen donors too


def _find_hydrogen_acceptors(topology):
    pass


class _Atom:
    def __init__(self, topology):
        if topology.n_atoms != 1:
            raise AttributeError('expecting 1 atom')
        self.topology = topology





# import molecular.flags as flags
import functools
import warnings


__all__ = [
    'deprecated',
    'experimental'
]


# Decorator to indicate if deprecated
# noinspection PyDeprecation
def deprecated(function=None):
    @functools.wraps(function)
    def execute_function(*args, **kwargs):
        warnings.warn('will be deprecated', DeprecationWarning)
        return function(*args, **kwargs)
    return execute_function


# Decorator to indicate if still in development
# TODO this should probably be moved out of __init__
# def experimental(f=None):
#     # Wrapper to run function
#     def execute_function(f):
#         @functools.wraps(f)
#         def wrapper(*args, **kwargs):
#             use_experimental = flags.use_experimental
#             # if 'use_experimental' in kwargs:
#             #     use_experimental = kwargs['use_experimental']
#             #     del kwargs['use_experimental']
#
#             # If false, then we cannot use this function
#             if not use_experimental:
#                 raise PermissionError('cannot use function because it is experimental')
#
#             return f(*args, **kwargs)
#         return wrapper
#
#     if f is None:
#         return execute_function
#     else:
#         return execute_function(f)

# Class to provide experimental warning
class ExperimentalWarning(UserWarning):
    pass


# Decorator to indicate if experimental
def experimental(obj=None):
    """
    Decorator to indicate that a class or function is experimental.

    Parameters
    ----------
    obj : None, class, or function

    Returns
    -------
    obj executed
    """

    @functools.wraps(obj)
    def execute(*args, **kwargs):
        warnings.warn('%s is experimental' % obj.__name__, ExperimentalWarning)
        return obj(*args, **kwargs)
    return execute



"""
geometry.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""


import numpy as np
from scipy.spatial.distance import squareform


# Compute angle between three points
# TODO introduce signed angles
def angle(a, b, c=None, method='atan2', degrees=False):
    r"""
    Compute the angle between two vectors or three vertices.

    If method = 'atan2',

    ..math :: \theta = atan2(norm(cross(a, b)), dot(a, b))

    If method = 'acos',

    ..math :: \theta = acos \frac{a \dot b}{|a| |b|}


    Parameters
    ----------
    a, b, c : ArrayLike
        Vectors with same dimensionality. Both `a` and `b` must be provided. If `c` is provided, all three are taken as
        vertices. Otherwise, `a` and `b` are assumed to be Euclidean vectors.
    method : str
        Method to compute angle, see :func:`~vangle` for options.
    degrees : bool
        Should the output be in degrees? (Default: False)

    Returns
    -------
    float or numpy.ndarray
        Angle between vectors
    """

    # If c is provided, compute Euclidean vectors
    # noinspection DuplicatedCode
    if c is not None:
        a = vector(a, b)
        b = vector(c, b)

    # Vectors must have at least 2 dimensions
    n_dim = _get_dimensions(a, b)
    if np.min(n_dim) < 2:
        raise AttributeError('angles can only be computed with 2D or 3D vectors (n_dim = %s)' % n_dim)

    # Coerce to a 2D form
    a, b, needs_ravel = _coerce_to_2d(a, b)

    # Format the method
    method = str(method).lower()

    # Compute the angle
    if method == 'atan2':
        cross_ = cross(a, b)
        if _has_dimensions(a, b, n_dim=2):
            if a.ndim == 2 and b.ndim == 2:
                cross_ = cross_.reshape(-1, 1)
            else:
                cross_ = [cross_]
        result = np.arctan2(norm(cross_), dot(a, b))

    elif method == 'acos':
        result = np.arccos(cos_angle(a, b))

    else:
        raise AttributeError('method %s unknown' % method)

    # Convert from radians to degrees?
    if degrees:
        result = np.rad2deg(result)

    # Return
    return _array_result(result, needs_ravel)


# Compute the gradient of an angle
# noinspection DuplicatedCode
def angle_gradient(a, b, c=None):
    """
    Gradient of angle with respect to vertices.

    Parameters
    ----------
    a, b, c : ArrayLike
        Cartesian coordinates.

    Returns
    -------
    """

    # If c is provided, compute Euclidean vectors
    if c is not None:
        a = vector(a, b)
        b = vector(c, b)

    # Vectors must have at least 2 dimensions
    n_dim = _get_dimensions(a, b)
    if np.min(n_dim) < 2:
        raise AttributeError('angles can only be computed with 2D or 3D vectors (n_dim = %s)' % n_dim)

    # Coerce to a 2D form
    a, b, needs_ravel = _coerce_to_2d(a, b)

    # Compute angle and factors
    cos_theta = cos_angle(a, b).reshape(-1, 1)
    inv_sin_theta = (1. / np.sqrt(1. - cos_theta * cos_theta)).reshape(-1, 1)
    inv_a_norm = (1. / norm(a)).reshape(-1, 1)
    inv_b_norm = (1. / norm(b)).reshape(-1, 1)
    a_unit = a * inv_a_norm
    b_unit = b * inv_b_norm

    # Compute gradients (here, gradient_a, gradient_b, and gradient_c stand for vertices)
    gradient_a = inv_sin_theta * inv_a_norm * (a_unit * cos_theta - a_unit)
    gradient_c = inv_sin_theta * inv_b_norm * (b_unit * cos_theta - b_unit)
    gradient_b = -1. * (gradient_a + gradient_c)

    # Return
    return np.array([
        _array_result(gradient_a, needs_ravel),
        _array_result(gradient_b, needs_ravel),
        _array_result(gradient_c, needs_ravel)
    ])


# Convert Cartesian to polar coordinates
def cartesian_to_polar(a):
    """
    Convert Cartesian to polar coordinates.

    Parameters
    ----------
    a : ArrayLike
        Cartesian coordinates.

    Returns
    -------
    numpy.ndarray
        Polar coordinates in same shape as `a`
    """

    # Coerce input
    a, needs_ravel = _coerce_to_2d(a)

    n_dim = a.shape[1]
    if n_dim in (2, 3):
        r = norm(a)
        result = [r, np.arctan2(a[:, 1], a[:, 0])]
        if n_dim == 3:
            result.append(np.arccos(a[:, 2] / r))

    else:
        raise AttributeError('cannot compute for %s dimensions' % n_dim)

    return _array_result(np.array(result).T, needs_ravel)


def condensed_to_square(a):
    return squareform(a)


# Compute angle between three points
def cos_angle(a, b, c=None):
    r"""
    Compute cosine of the angle between three points :math:`\angle ABC`.

    Parameters
    ----------
    a, b, c : ArrayLike
        Vectors

    Returns
    -------
    float or numpy.ndarray
        Cosine angle of points
    """

    if c is not None:
        a, b = vector(a, b), vector(c, b)

    # Coerce
    a, b, needs_ravel = _coerce_to_2d(a, b)

    # Return
    # TODO should cos_angle=np.clip(dot_prod/(len1*len2),-1.,1.)
    return _array_result(dot(a, b) / (norm(a) * norm(b)), needs_ravel)


# Cross product
def cross(a, b):
    """
    Compute cross product between vectors `a` and `b`. Points to :func:`numpy.cross`

    Parameters
    ----------
    a, b : ArrayLike
        Vectors

    Returns
    -------
    float or numpy.ndarray
        Cross product of u and v
    """

    return np.cross(a, b)


# Dihedral angle between 4 points
def dihedral(a, b, c, d=None):
    """
    Compute dihedral angle between four points.

    TODO add equation for sanity checking

    Parameters
    ----------
    a, b, c, d : ArrayLike
        Cartesian coordinates. If only `a`, `b`, and `c` are supplied, they are taken as vectors. If `d` is included,
        they are taken as vertices.

    Returns
    -------
    float or numpy.ndarray
        Dihedral angle
    """

    if d is not None:
        a, b, c = vector(b, a), vector(c, b), vector(d, c)

    _check_dimensions(a, b, c, n_dim=3)

    u = cross(a, b)
    v = cross(b, c)

    return angle(u, v, method='acos')


# Compute the distance between two vectors
def distance(a, b=None):
    """
    Compute the Euclidean distance between two vectors.

    Parameters
    ----------
    a : ArrayLike
        Vector
    b : ArrayLike
        (Optional) Vector. If None, `b` is set to 0 and the distance to the origin is computed.

    Returns
    -------
    float
        Distance
    """

    # Coerce
    a, needs_ravel = _coerce_to_2d(a)

    # If y is not supplied, set to zeros; then coerce
    if b is None:
        b = np.zeros(a.shape)
    b, _ = _coerce_to_2d(b)

    # Return distance
    return _array_result(np.sqrt(np.sum(np.square(vector(a, b)), axis=1)), needs_ravel)


# Dot product
def dot(a, b, axis=1):
    """
    Compute vector dot product.

    Parameters
    ----------
    a, b : ArrayLike
        Vectors
    axis : int
        Axis to apply sum over.

    Returns
    -------
    float or numpy.ndarray
        Vector dot product
    """

    a, b, needs_ravel = _coerce_to_2d(a, b)

    return _array_result(np.sum(np.multiply(a, b), axis=axis), needs_ravel)


# Normed vector
def norm(a):
    r"""
    Compute vector norm.

    .. math :: |a| = \sqrt{a_x^2 + a_y^2 + a_z^2}

    Parameters
    ----------
    a : ArrayLike
        Vector

    Returns
    -------
    float or numpy.ndarray
        Normed vector
    """

    # a, needs_ravel = _coerce_to_2d(a)
    #
    # return _array_result(np.linalg.norm(a, axis=1), needs_ravel)
    return distance(a)


# Compute the normal between three points
# FIXME is this confusing? normal = cross
def normal(a, b, c=None):
    """
    Compute normal vector between three points.

    Parameters
    ----------
    a, b, c : ArrayLike
        Cartesian coordinates.

    Returns
    -------
    numpy.ndarray
        Vector normal
    """

    if c is not None:
        a, b = vector(a, b), vector(c, b)

    return cross(a, b)


# Polar to cartesian coordinates
def polar_to_cartesian(a):
    """
    Convert polar to Cartesian coordinates.

    Parameters
    ----------
    a : ArrayLike
        Polar coordinates

    Returns
    -------
    float or numpy.ndarray
        Cartesian coordinates, same shape as `a`
    """

    #
    # a = np.array(a).reshape(-1)

    # Coerce
    a, needs_ravel = _coerce_to_2d(a)
    # if b is not None:
    #     b = np.array(b).reshape(-1)
    #     b, _ = _coerce_to_2d(b)
    #     a = np.hstack([a, b])

    n_dim = a.shape[1]
    if n_dim in (2, 3):
        result = np.array([np.cos(a[:, 1]), np.sin(a[:, 1])])
        if n_dim == 3:
            result *= np.sin(a[:, 2])
            result = np.vstack([result, np.cos(a[:, 2])])
        result *= a[:, 0]

    else:
        raise AttributeError('cannot compute for %s dimensions' % n_dim)

    return _array_result(np.array(result).T, needs_ravel)


# Create unit vector
def unit_vector(a):
    r"""
    Compute unit vector.

    .. math :: \^{a} = \frac{a}{|a|}

    Parameters
    ----------
    a : ArrayLike
        Vector

    Returns
    -------
    numpy.ndarray
        Unit vector
    """

    a, needs_ravel = _coerce_to_2d(a)
    return _array_result(a / norm(a).reshape(-1, 1), needs_ravel)


#
def vangle_gradient(u, v):
    """

    Parameters
    ----------
    u
    v

    Returns
    -------

    """

    cos_theta = cos_angle(u, v)
    inv_sin_theta = 1. / np.sqrt(1. - cos_theta * cos_theta)
    inv_u_norm = 1. / norm(u)
    inv_v_norm = 1. / norm(v)
    u_unit = u * inv_u_norm
    v_unit = v * inv_v_norm

    gradient_a = inv_sin_theta * inv_u_norm * (u_unit * cos_theta - v_unit)
    gradient_c = inv_sin_theta * inv_v_norm * (v_unit * cos_theta - u_unit)
    gradient_b = -1. * (gradient_a + gradient_c)

    print(gradient_a)


# Compute vector between 2 sets of points
def vector(a, b, normed=False):
    """
    Compute vector between two sets of points.

    .. math:: vector = a - b

    Parameters
    ----------
    a, b : ArrayLike
        Cartesian coordinates.
    normed : bool
        Should the unit vector be computed? (Default: False)

    Returns
    -------
    numpy.ndarray
        Vector between `a` and `b`.
    """

    # Coerce input
    a, b, needs_ravel = _coerce_to_2d(a, b)

    v = np.subtract(a, b)
    if normed:
        v /= norm(v).reshape(-1, 1)

    return _array_result(v, needs_ravel)


# Helper function to ravel array result
def _array_result(a, needs_ravel=False):
    if needs_ravel:
        if a.ndim > 1:
            a = a.ravel()
        else:
            a = a[0]
    return a


def _check_dimensions(*args, n_dim=3):
    if not _has_dimensions(*args, n_dim=n_dim):
        raise AttributeError('must be %sD' % n_dim)


def _coerce_to_2d(a, *args):
    # Coerce a to 2d
    a = np.array(a)
    needs_ravel = False
    if a.ndim == 1:
        needs_ravel = True
        a = a.reshape(1, -1)
    result = [a]

    # Do the same for kwargs
    for arg in args:
        b, _ = _coerce_to_2d(arg)
        if a.shape != b.shape:
            raise AttributeError('vectors must be same shapes')
        result.append(b)

    # Add in needs_ravel flag
    result.append(needs_ravel)

    # Return
    return result


def _get_dimensions(*args, n_dim=3):
    result = []
    for arg in args:
        arg = np.array(arg)
        n_dim = 1
        if arg.ndim > 1:
            n_dim = arg.shape[1]
        elif arg.ndim == 1:
            n_dim = arg.shape[0]
        result.append(n_dim)
    return result


# Helper function to check dimensionality
def _has_dimensions(*args, n_dim=3):
    result = []
    for arg in args:
        arg = np.array(arg)
        result.append((arg.ndim > 1 and arg.shape[1] == n_dim) or (arg.ndim == 1 and arg.shape[0] == n_dim))
    return all(result)





"""
rmsd.py
written in Python3
author: C. Lockhart

>>> from molecular import read_pdb, rmsd
>>> trajectory = read_pdb('trajectory1.pdb')
>>> first_structure = trajectory[0]
>>> rmsd = rmsd(trajectory, first_structure)

"""

# noinspection PyProtectedMember
from molecular.transform.transform import _fit
from itertools import product
import numpy as np


# Compute the RMSD between 2 trajectories
def rmsd(a, b=None, paired=False, fit=True):
    """
    Compute the RMSD.

    If only `a` is provided, then the RMSD is computed between all structures in the Trajectory.
    Otherwise, if `a` and `b` are provided, we compute the RMSD between all structures in `a` and all structures in `b`.


    Parameters
    ----------
    a : molecular.Trajectory
    b : molecular.Trajectory
    paired : bool
        Are a and b paired? That is, should we compute RMSD between (a[0], b[0]), (a[1], b[1]), etc.? (Default: False)
    fit : bool
        Should structures be fit before RMSD is computed? (Default: True)

    Returns
    -------


    See Also
    --------
    http://manual.gromacs.org/documentation/current/onlinehelp/gmx-rms.html
    """

    # If a is None, then select b
    if a is None:
        b = a

    # Number of atoms between a and b must be identical
    if a.n_atoms != b.n_atoms:
        raise AttributeError('a and b must have same number of atoms')

    # Get coordinates
    a_xyz = a.xyz
    b_xyz = b.xyz

    # Compute paired?
    if paired:
        if len(a) != len(b):
            raise AttributeError('cannot compute paired RMSD because a and b have different number of structures')
        # iterable = zip(range(a.n_structures), range(b.n_structures))
        a_index = range(a.n_structures)
        b_index = range(b.n_structures)

    # Otherwise, compute RMSD taking a x b
    else:
        # FIXME the problem with this that it's memory-intensive to replicate ...
        # iterable = product(range(a.n_structures), range(b.n_structures))
        ab_product = np.array(list(product(range(a.n_structures), range(b.n_structures))))
        a_index, b_index = ab_product[:, 0], ab_product[:, 1]

    # Expand a_xyz and b_xyz so they have the same dimensions and are "paired"
    a_xyz = a_xyz[a_index, :, :]
    b_xyz = b_xyz[b_index, :, :]

    # Should we fit the two structures?
    if fit:
        b_xyz = _fit(a_xyz, b_xyz)

    # Compute the difference between a and b
    # diff = a_xyz[a_index, :, :] - b_xyz[b_index, :, :]

    # Actually Compute RMSD
    result = np.sqrt(np.sum(np.square(a_xyz - b_xyz), axis=(1, 2)) / a.n_atoms)
    # result = np.zeros((a.n_structures, b.n_structures))
    # for i, j in iterable:
    #     result[i, j] = np.sqrt(np.mean((a_xyz[i, :, :] - b_xyz[j, :, :]) ** 2))

    # Pivot into wide form
    # https://stackoverflow.com/questions/17028329/python-create-a-pivot-table
    # rows, row_pos = np.unique(a_index, return_inverse=True)
    # cols, col_pos = np.unique(b_index, return_inverse=True)
    # pivot_table = np.zeros((len(rows), len(cols)))
    # pivot_table[row_pos, col_pos] = result
    result = result.reshape(a.n_structures, b.n_structures)

    # Return
    return result

#
# def _rmsd(a, b, paired=False, fit=True):
#     """
#
#     Parameters
#     ----------
#     a
#     b
#
#     Returns
#     -------
#
#     """
#
#     # Get the number of elements in a and b
#     n_a = len(a)
#     n_b = len(b)
#
#     a_xyz = a.xyz
#     b_xyz = b.xyz
#
#     # Should we fit the two structures?
#     if fit:
#         b_xyz = _fit(a, b)
#
#     result = np.zeros((n_a, n_b))
#     if not paired:
#         iterable = product(range(n_a), range(n_b))
#     else:
#         iterable = zip(range(n_a), range(n_b))
#     for i, j in iterable:
#         result[i, j] = np.sqrt(np.mean((a_xyz[i, :, :] - b_xyz[j, :, :]) ** 2))
#
#     return result



"""
geometry.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from .geometry import *

__all__ = [
    'angle',
    'angle_gradient',
    'cartesian_to_polar',
    'condensed_to_square',
    'cos_angle',
    'cross',
    'dihedral',
    'distance',
    'dot',
    'norm',
    'normal',
    'polar_to_cartesian',
    'unit_vector',
    'vector',
]





import yaml
import os
import pandas as pd

include_dir = os.path.abspath(__file__ + '../../../../_include')


# Read relative SASA
def get_relative_sasa(sequence=None, which='Tien'):
    """
    Get relative SASA for sequence. Relative SASA is effectively the maximum SASA allowed for a residue type. It's
    typically computed in the form Gly-X-Gly, where X is the residue of interest.

    Rose


    Tien
    https://journals.plos.org/plosone/article/file?id=10.1371/journal.pone.0080635&type=printable


    ala-x-ala???
    https://doi.org/10.1016/0022-2836(91)90027-4

    Parameters
    ----------
    sequence : np.ndarray or None
        If None, relative SASA for all 20 common amino acids are returned.
    which : str
        Which version of relative SASA should we use? Choices include Rose, Tien... (Default: Tien)

    Returns
    -------
    pandas.Series
        Relative SASA.
    """

    # Read in normalizing SASA values
    with open(os.path.join(include_dir, 'protein', 'relative_sasa.yml'), 'r') as stream:
        data = yaml.safe_load(stream.read())

    # Get specific normalize dataset
    data = data[which]

    # Convert to Series
    relative_sasa = pd.Series(data).rename('relative_sasa')

    # Result
    result = relative_sasa
    if sequence is not None:
        result = relative_sasa.loc[sequence]

    # Return
    return result


class SurfaceArea:
    pass


if __name__ == '__main__':
    print(get_relative_sasa())



"""
distance.py

language: python
version: 3.x
author: C. Lockhart <chris@lockhartlab.org>
"""


import numpy as np


def distance(a, b):
    """
    Compute the distance between two Trajectory instances.

    Parameters
    ----------
    a, b : Trajectory
        Two trajectories. Must have same dimensions.

    Returns
    -------
    numpy.ndarray
        Distance between every frame in the trajectory.
    """

    return np.sqrt(np.sum(np.square(a.xyz - b.xyz), axis=(1, 2)))




import numpy as np


def reptemp(temp_min=300, temp_max=440, n_replicas=40, mode='geometric'):
    """
    Geometric = the ratio from one temperature to the next is kept constant

    .. math :: T_i = T_1 \frac{T_R}{T_1}^{\frac{i-1}{R-1}}

    .. math :: log T_i - log T_1 = \frac{i-1}{R-1}(log


    Parameters
    ----------
    temp_min : float
    temp_max : float
    n_replicas : int
    mode : str

    Returns
    -------

    """

    if mode != 'geometric':
        raise AttributeError('only supports geometric')

    return temp_min * np.power(temp_max / temp_min, np.arange(n_replicas) / (n_replicas - 1.), dtype=np.float64)


if __name__ == '__main__':
    print(reptemp(300, 440, 40))



"""
rtf.py

description: Read RTF from CHARMM format.
language: Python 3.x
author: C. Lockhart <chris@lockhartlab.org>
"""

"""
Topology
Trajectory

ForceField
"""


class ForceField:
    def __init__(self):
        # self.atom_types = {}
        self.residues = {}

    def __setitem__(self, key, value):
        self.residues[key] = value

    def __getitem__(self, item):
        return self.residues[item]


class Residue:
    def __init__(self, name, charge):
        self.name = str(name)
        self.charge = float(charge)
        self.atoms = {}
        self.bonds = []
        self.angles = []
        self.dihedrals = []
        self.impropers = []
        self.crossterms = []
        self.donors = []
        self.acceptors = []

    def __add__(self, atom):
        assert isinstance(atom, Atom)
        self.atoms.append(atom)
        return self

    def __repr__(self):
        s = 'ATOM    MASS    CHARGE\n'
        for atom_name, atom in self.atoms.items():
            s += f'{atom_name} {atom.mass} {atom.charge}\n'
        return s


class Patch:
    def __init__(self):
        pass


class Atom:
    # noinspection PyShadowingBuiltins
    def __init__(self, type, charge, mass):
        self.type = type
        self.charge = charge
        self.mass = mass


def read_rtf(fname, debug=False):
    """

    Parameters
    ----------
    fname

    Returns
    -------

    """

    # Read in file
    with open(fname, 'r') as stream:
        data = stream.readlines()
        # if debug:
        #     print(data)

    # Create empty ForceField
    forcefield = ForceField()

    #
    mass = {}

    # Read line-by-line
    section = 0
    for line in data:
        # print(line)
        # Skip lines with comments or blank
        if line[0] == '!' or line == '':
            continue

        # If line starts with an asterisk, print it out
        if line[0] == '*':
            print(line.strip())
            continue

        # The first line whose first character without an asterisk tells us we're ready to start
        if section == 0 and line[0] != '*':
            section = 1

        # Expect to read force field version number
        if section == 1:
            print(line.strip())
            section = 2
            continue

        # Divide line into columns
        col = line.split('!')[0].split()
        columns = col

        # If we have no columns, skip
        if len(columns) == 0:
            continue

        # Read MASS section
        if section == 2 and columns[0] == 'MASS':
            assert len(columns) >= 4
            atom_type = columns[2]
            mass[atom_type] = float(col[3].split('!')[0])
            continue
        elif section == 2 and columns[0] != 'MASS':
            section = 3

        # Read RESI section
        if section == 3 and line[:4] == 'RESI':
            residue_name = col[1]
            residue_charge = col[2].split('!')[0]
            forcefield.residues[residue_name] = Residue(name=residue_name, charge=residue_charge)
            continue
        elif section == 3 and line[:4] == 'ATOM':
            atom_name = col[1]
            atom_type = col[2]
            atom_charge = col[3].split('!')[0]
            atom_mass = mass.get(atom_type, 0.)
            # if atom_type not in forcefield.atom_types:
            #     forcefield.atom_types[atom_type] = Atom(type=atom_type, charge=atom_charge, mass=atom_mass)
            # noinspection PyUnboundLocalVariable
            # forcefield.residues[residue_name].atoms[atom_name] = forcefield.atom_types[atom_type]
            forcefield.residues[residue_name].atoms[atom_name] = Atom(type=atom_type, charge=atom_charge,
                                                                      mass=atom_mass)
            continue
        elif section == 3 and (line[:4] in ['BOND', 'DOUB', 'TRIP']):
            bond_type = 'bond'
            if line[:4] == 'DOUB':
                bond_type = 'double'
            elif line[:4] == 'TRIP':
                bond_type = 'triple'
            for i in range(1, len(col), 2):
                atom_name0 = col[i].split('!')[0]
                atom_name1 = col[i + 1].split('!')[0]
                forcefield.residues[residue_name].bonds.append([atom_name0, atom_name1, bond_type])
        elif section == 3 and line[:4] == 'ANGL':
            for i in range(1, len(col), 3):
                atom_name0 = col[i].split('!')[0]
                atom_name1 = col[i + 1].split('!')[0]
                atom_name2 = col[i + 2].split('!')[0]
                forcefield.residues[residue_name].angles.append([atom_name0, atom_name1, atom_name2])
        elif section == 3 and line[:4] == 'DIHE':
            for i in range(1, len(col), 4):
                atom_name0 = col[i].split('!')[0]
                atom_name1 = col[i + 1].split('!')[0]
                atom_name2 = col[i + 2].split('!')[0]
                atom_name3 = col[i + 3].split('!')[0]
                forcefield.residues[residue_name].dihedrals.append([atom_name0, atom_name1, atom_name2, atom_name3])
        elif section == 3 and line[:4] == 'IMPR':
            for i in range(1, len(col), 4):
                atom_name0 = col[i].split('!')[0]
                atom_name1 = col[i + 1].split('!')[0]
                atom_name2 = col[i + 2].split('!')[0]
                atom_name3 = col[i + 3].split('!')[0]
                forcefield.residues[residue_name].impropers.append([atom_name0, atom_name1, atom_name2, atom_name3])
        elif section == 3 and line[:4] == 'CMAP':
            for i in range(1, len(col), 8):
                atom_name0 = col[i].split('!')[0]
                atom_name1 = col[i + 1].split('!')[0]
                atom_name2 = col[i + 2].split('!')[0]
                atom_name3 = col[i + 3].split('!')[0]
                atom_name4 = col[i + 4].split('!')[0]
                atom_name5 = col[i + 5].split('!')[0]
                atom_name6 = col[i + 6].split('!')[0]
                atom_name7 = col[i + 7].split('!')[0]
                forcefield.residues[residue_name].crossterms.append([atom_name0, atom_name1, atom_name2, atom_name3,
                                                                     atom_name4, atom_name5, atom_name6, atom_name7])
        elif section == 3 and line[:4] == 'DONO':
            for i in range(1, len(col), 1):
                atom_name = col[i].split('!')[0]
                forcefield.residues[residue_name].donors.append(atom_name)
        elif section == 3 and line[:4] == 'ACCE':
            for i in range(1, len(col), 1):
                atom_name = col[i].split('!')[0]
                forcefield.residues[residue_name].acceptors.append(atom_name)

        # TODO read PRES section

    # We're done
    return forcefield


if __name__ == '__main__':
    import os

    include_dir = os.path.abspath(os.path.join(__file__, '../../../samples'))

    rtf = read_rtf(os.path.join(include_dir, 'top_all27_prot_lipid.inp'), debug=True)







"""
pdb.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.analysis.rmsd import rmsd
# noinspection PyProtectedMember
from molecular.io.read import _read_pdb

import numpy as np
from urllib.request import urlopen


# noinspection PyPep8Naming,SpellCheckingInspection
def RCSB(pdb_id, most_common=False):
    """
    Download PDB from RCSB.

    Parameters
    ----------
    pdb_id : str
        Reference ID for the PDB structure.
    most_common : bool
        Extract only the PDB structure with the lowest internal RMSD.

    Returns
    -------
    molecular.Trajectory
        Instance of Trajectory object.
    """

    # Read RCSB and return molecular Trajectory
    result = _read_pdb(urlopen('http://files.rcsb.org/download/' + pdb_id + '.pdb').read().decode('utf-8'))

    # If most_common, compute the pairwise RMSD and select structure structure with minimum average value
    if most_common:
        r = rmsd(result, paired=False, fit=True)
        i = np.argmin(r.mean(axis=0))
        result = result[i]

    # Return
    return result




"""
read.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""

from molecular.core import Topology, Trajectory

import numpy as np
# from numpy.lib.recfunctions import drop_fields, structured_to_unstructured
import pandas as pd
import re
from typelike import ArrayLike


# Read PDB
# TODO currently the only backend will by pandas; in future, expand to Cython or C or Fortran backend
def read_pdb(filename, backend='python'):
    """
    Read PDB file and return Trajectory

    PDB format can be described in depth at `<http://www.wwpdb.org/documentation/file-format>`_.

    Parameters
    ----------
    filename : str
        Name of PDB file to be read
    backend : str
        (Default: 'pandas').

    Returns
    -------
    molecular.Trajectory
        Trajectory of PDB
    """

    # Make sure we know we're using the pandas backend
    if backend.lower() != 'python':
        raise AttributeError('only python backend presently supported')

    # Open file, read in all records
    with open(filename, 'r') as stream:
        records = stream.read()
        # records = _atom_reader(buffer)

    # Return
    return _read_pdb(records)


def read_peptide_sequence(residues):
    """
    Create trajectory from an amino acid sequence.

    Parameters
    ----------
    residues : str or ArrayLike
        List of amino acids.

    Returns
    -------
    molecular.Trajectory
        Instance of Trajectory.
    """

    # Convert to list if not already list
    if not isinstance(residues, ArrayLike):
        residues = list(residues)

    # Make sure we're in the right format
    letter_to_code = {
        'A': 'ALA',
        'R': 'ARG',
        'N': 'ASN',
        'D': 'ASP',
        'C': 'CYS',
        'Q': 'GLN',
        'E': 'GLU',
        'G': 'GLY',
        'H': 'HIS',
        'I': 'ILE',
        'L': 'LEU',
        'K': 'LYS',
        'M': 'MET',
        'F': 'PHE',
        'P': 'PRO',
        'S': 'SER',
        'T': 'THR',
        'W': 'TRP',
        'Y': 'TYR',
        'V': 'VAL'
    }
    residues = [letter_to_code.get(residue, residue) for residue in residues]

    # Create Trajectory
    topology = Topology()
    topology._data = pd.DataFrame({'residue_id': np.arange(len(residues)), 'residue': residues})

    # Create Trajectory
    trajectory = Trajectory(xyz=np.zeros((1, len(residues), 3)), topology=topology)

    # Return
    return trajectory


def _read_pdb(records):
    # Filter out CRYST1 records
    cryst = re.sub(r'^(?!CRYST1).*$', '', records, flags=re.MULTILINE).lstrip()
    n_structures = len(re.split('\WCRYST1', cryst))

    # Filter out atom records
    # TODO this will be slow for large PDB files; perhaps move to Cython or C backend
    atoms = re.sub(r'^(?!ATOM).*$', '', records, flags=re.MULTILINE).replace('\n\n', '\n').lstrip()

    # Sections of PDB
    sections = np.array([
        (6, 'record', '<U6'),
        (5, 'atom_id', 'int'),
        (5, 'atom', '<U5'),
        (5, 'residue', '<U5'),
        (1, 'chain', '<U1'),
        (4, 'residue_id', 'int'),
        (4, 'blank', '<U1'),
        (8, 'x', 'float'),
        (8, 'y', 'float'),
        (8, 'z', 'float'),
        (6, 'alpha', 'float'),
        (6, 'beta', 'float'),
        (10, 'segment', '<U9'),
        (2, 'element', '<U2')
    ], dtype=[('width', 'i1'), ('column', '<U10'), ('type', '<U10')])

    # Read in
    data = np.genfromtxt(atoms.split('\n'), delimiter=sections['width'], dtype=sections['type'], autostrip=True)
    # data.dtype.names = sections['column']
    data = pd.DataFrame(data.tolist(), columns=sections['column'])

    # Drop extraneous columns
    # data = drop_fields(data, 'blank')
    data = data.drop(columns='blank')

    # TODO this should also be done for residue_id probably
    # If the PDB starts at atom_id = 1, change to 0-index
    # if data['atom_id'].min() == 1:
    #     data['atom_id'] -= 1

    # Determine number of structures in PDB
    # atom_id = data['atom_id'].values
    # pos_atom_id = atom_id[atom_id >= 0]
    # _, atom_counts = np.unique(pos_atom_id, return_counts=True)
    # n_structures = np.unique(atom_counts)
    # # n_structures = data.pivot_table(index='atom_id', values='record', aggfunc='count')['record'].unique()
    # if len(n_structures) != 1:
    #     raise AttributeError('inconsistent record counts in PDB, %s' % n_structures)
    #     # raise AttributeError('inconsistent record counts in PDB')
    # n_structures = n_structures[0]

    # Separate out dynamic columns for Trajectory and static Topology data
    dynamical_columns = ['x', 'y', 'z']
    # static_columns = [column for column in data.dtype.names if column not in dynamical_columns]
    static_columns = [column for column in data.columns if column not in dynamical_columns]

    # Renumber atom_id
    if np.mod(len(data), n_structures) != 0:
        raise AttributeError('len(data) must be evenly divisible by n_structures, (%s, %s)' % (len(data), n_structures))
    data['atom_id'] = np.array(np.tile(np.arange(1, len(data) / n_structures + 1), n_structures), dtype='int')

    # Create Topology first
    # TODO what happens when alpha and beta differ by structures? Should these be stored in Trajectory?
    data['alpha'] = 0.
    data['beta'] = 0.
    # topology = Topology(np.unique(data[static_columns]))
    topology = Topology(data[static_columns].drop_duplicates())

    # Next create Trajectory (the result)
    # n_atoms = len(np.unique(data['atom_id'].values))
    n_atoms = data['atom_id'].nunique()
    # result = Trajectory(structured_to_unstructured(data[dynamical_columns]).reshape(n_structures, n_atoms, 3),
    #                     topology=topology)
    result = Trajectory(data[dynamical_columns].values.reshape(n_structures, n_atoms, 3), topology=topology)

    # Return
    return result


# @jit(nopython=False)
# def _atom_reader(buffer):
#     result = ''
#     for line in buffer.readlines():
#         if line[:4] == 'ATOM':
#             result += line
#     return result

# Read DCD
def read_dcd(filename, topology=None):
    """
    Read in DCD file with `filename`. This function is partially based off James Phillips' code MDTools that is no 
    longer in development. See http://www.ks.uiuc.edu/Development/MDTools/Python/

    https://www.ks.uiuc.edu/Research/namd/wiki/index.cgi?ReadingDCDinFortran
    
    Parameters
    ----------
    filename : str
        Name of DCD file.
    topology : molecular.Topology
        (Optional) Topology file to load for coordinates.

    Returns
    -------
    molecular.Trajectory
        Instance of Trajectory object.
    """

    # Open binary DCD file for reading
    stream = open(filename, 'rb')

    # Read first part of header and choose endianness
    # File is Fortran unformatted, which tells us the number of bytes
    # in each part. This first section should contain 84 bytes.
    endian = '>'
    n_byte = np.ndarray((1,), endian + 'i', stream.read(4))[0]
    title = np.ndarray((4,), endian + 'B', stream.read(4)).tostring().decode('ASCII')
    if n_byte != 84 or title != 'CORD':
        endian = '<'
        stream.close()
        stream = open(filename, 'rb')
        n_byte = np.ndarray((1,), endian + 'i', stream.read(4))[0]
        title = np.ndarray((4,), endian + 'B', stream.read(4)).tostring().decode('ASCII')
        if n_byte != 84 or title != 'CORD':
            raise IOError('cannot read DCD header')

    # Continue reading rest of header
    n_structures = np.ndarray((1,), endian + 'i', stream.read(4))[0]
    stream.read(28)
    fixed = np.ndarray((1,), endian + 'i', stream.read(4))[0]
    if fixed != 0:
        raise IOError('failed reading DCD file')
    stream.read(44)
    if np.ndarray((1,), endian + 'i', stream.read(4))[0] != n_byte:
        raise IOError('failed reading DCD file')

    # Read next section, which contains 164 bytes
    n_byte = np.ndarray((1,), endian + 'i', stream.read(4))[0]
    if n_byte != 164:
        raise IOError('failed reading DCD file')
    n_titles = np.ndarray((1,), endian + 'i', stream.read(4))[0]
    for i in range(n_titles):
        title = np.ndarray((80,), endian + 'B', stream.read(80)).tostring().decode('ASCII')
    if np.ndarray((1,), endian + 'i', stream.read(4))[0] != n_byte:
        raise IOError('failed reading DCD file')

    # Read next section, which contains 4 bytes
    n_byte = np.ndarray((1,), endian + 'i', stream.read(4))[0]
    if n_byte != 4:
        raise IOError('failed reading DCD file')
    n_atoms = np.ndarray((1,), endian + 'i', stream.read(4))[0]
    if n_atoms == 0:
        raise ValueError('structure empty')
    if np.ndarray((1,), endian + 'i', stream.read(4))[0] != n_byte:
        raise IOError('failed reading DCD file')

    # Get information from structures
    buffer = stream.read()

    # Box information
    n_byte = np.ndarray((n_structures,), endian + 'i', buffer, strides=(80 + 12 * n_atoms))
    if any(n_byte != 48):
        raise IOError('failed reading DCD file')
    box_x = np.ndarray((n_structures,), endian + 'f8', buffer, offset=4, strides=(80 + 12 * n_atoms))
    box_y = np.ndarray((n_structures,), endian + 'f8', buffer, offset=20, strides=(80 + 12 * n_atoms))
    box_z = np.ndarray((n_structures,), endian + 'f8', buffer, offset=44, strides=(80 + 12 * n_atoms))
    if any(np.ndarray((n_structures,), endian + 'i', buffer, offset=52, strides=(80 + 12 * n_atoms)) != n_byte):
        raise IOError('failed reading DCD file')

    # xyz
    # TODO there might be a way to clean this up
    # noinspection PyShadowingNames
    def _xyz(off1, off2, off3):
        n_byte = np.ndarray((n_structures,), endian + 'i', buffer, offset=off1, strides=(80 + 12 * n_atoms))
        if any(n_byte != 4 * n_atoms):
            raise IOError('failed reading DCD file')
        r = np.ndarray((n_structures, n_atoms), endian + 'f', buffer, offset=off2, strides=(80 + 12 * n_atoms, 4))
        if any(np.ndarray((n_structures,), endian + 'i', buffer, offset=off3, strides=(80 + 12 * n_atoms)) != n_byte):
            raise IOError('failed reading DCD file')
        return r

    x = _xyz(56 + 0 * n_atoms, 60 + 0 * n_atoms, 60 + 4 * n_atoms)
    y = _xyz(64 + 4 * n_atoms, 68 + 4 * n_atoms, 68 + 8 * n_atoms)
    z = _xyz(72 + 8 * n_atoms, 76 + 8 * n_atoms, 76 + 12 * n_atoms)

    # Create Trajectory and return
    return Trajectory(np.dstack([x, y, z]), box=np.vstack([box_x, box_y, box_z]).T, topology=topology)




from molecular.geometry import angle

import matplotlib.pyplot as plt
from matplotlib import patches
import numpy as np


def plot_vector_angle(u, v, color=(0, 0.5, 1)):
    plt.xkcd()
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot([0, 0], [0, 0])
    ax.arrow(0, 0, u[0], u[1], linewidth=2., head_width=0.05, head_length=0.03, fc=color, ec=color, zorder=2)
    ax.arrow(0, 0, v[0], v[1], linewidth=2., head_width=0.05, head_length=0.03, fc=color, ec=color, zorder=2)
    angles = 360. - np.array([angle(u, [1., 0.]), angle(v, [1., 0.])]) * 180. / np.pi
    angles[angles >= 360] -= 360
    angles_min = np.min(angles)
    angles_max = np.max(angles)
    if angles_max - angles_min > (360 - angles_max) + angles_min:
        theta1 = np.max(angles)
        theta2 = theta1 + (360 - angles_max) + angles_min
    else:
        theta1 = np.min(angles)
        theta2 = theta1 + angles_max - angles_min
    e1 = patches.Arc([0., 0.], 1., 1., angle=0., theta1=theta1, theta2=theta2, color=color, zorder=2)
    ax.add_patch(e1)
    ax.spines['left'].set_position('zero')
    ax.spines['right'].set_color('none')
    ax.spines['bottom'].set_position('zero')
    ax.spines['top'].set_color('none')
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])
    ax.set_aspect('equal')
    plt.show()


def heatmap(x, y):
    # plt.plot()
    # plt.imshow(, cmap='rainbow')
    # plt.show()
    pass




import numpy as np
from sympy.ntheory import factorint

from .replica_exchange import *

__all__ = [
    'get_pme_size',
    'reptemp',
]


# TODO move this to a better location
# TODO does this actually belong in the package namdtools?
def get_pme_size(a):
    """
    Given a box length, figure out the grid size for PME. We are looking for the PME size that's ~1 grid spacing, while
    with only small prime factors 2, 3, and 5.

    Parameters
    ----------
    a : numeric
        Size of box length.

    Returns
    -------
    int
        Size of PME grid length.
    """

    # If a is a decimal, round up
    a = int(np.ceil(a))

    # We want to find larger number that has only factors 2, 3, and 5
    # noinspection PyShadowingNames
    def _has_only_factors(a):
        factors = np.array(list(factorint(a).keys()))
        for factor in [2, 3, 5]:
            factors = np.delete(factors, np.where(factors == factor))
        return len(factors) == 0

    # Keep on increasing a until we find the right number
    while not _has_only_factors(a):
        a += 1

    # Return
    return a


if __name__ == '__main__':
    print(get_pme_size(125.))




class CompartmentalModel:
    def __init__(self, compartments, parameters):
        self._compartments = {}

    def __add__(self, other):
        self.add_compartment(other)

    def __getitem__(self, item):
        return self._compartments[item]

    def __setitem__(self, key, value):
        self._compartments[key] = value

    def add_compartment(self, compartment):
        if not isinstance(compartment, Compartment):
            raise AttributeError('must be Compartment')
        label = compartment.label
        if label in self._compartments:
            raise AttributeError('Compartment label already used')
        self._compartments[label] = compartment


class Compartment:
    def __init__(self, initial_value):
        self._initial_value = initial_value
        self._derivative = None

    def __sub__(self, other):
        pass

    @property
    def derivative(self):
        return self._derivative

    @derivative.setter
    def derivative(self, value):
        self._derivative = value

    @property
    def label(self):
        return self._label


class Parameter:
    def __init__(self, value):
        self._value = value

    @property
    def inverse(self):
        return -self._value


model = CompartmentalModel(
    compartments={
        'S': 1,
        'I': 1.27e-6,
        'R': 0,
    },
    parameters={
        'b': 0.5,
        'k': 0.33,
    }
)

# S, I, R = model.compartments
# b, k = model.parameters

S, I, R = Compartment(initial_value=1.), Compartment(initial_value=1.27e-6), Compartment(initial_value=0.)
b, k = Parameter(0.5), Parameter(0.33)

S.derivative = b.inverse * S * I
I.derivative = b * S * I - k * I
R.derivative = k * I



#
# model.add_equation('S = S - b * S * I')
# model.add_equation('I = I + b * S * I - k * I')
# model.add_equation('R = R + k * I')

model.integrate(500)

